/* $USAGI: base64.c,v 1.1 2004/07/05 16:42:50 yoshfuji Exp $ */
/*
 * Copyright (C) 2004 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * Author: Hideaki YOSHIFUJI <yoshfuji@linux-ipv6.org>
 */
#if BASE64_DEBUG
#include <stdio.h>
#include <stdlib.h>
#endif
#include <stdint.h>

#define	END	-1
#define	INV	-2
#define EOS	-3

int base64_table[] = {
	EOS,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,
	INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,
	INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV, 62,INV,INV,INV, 63,
	 52, 53, 54, 55, 56, 57, 58, 59, 60, 61,INV,INV,INV,END,INV,INV,
	INV,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
	 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,INV,INV,INV,INV,INV,
	INV, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
	 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51,INV,INV,INV,INV,INV,
	INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,
	INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,
	INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,
	INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,
	INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,
	INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,
	INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,
	INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,INV,
};

static int b64_decode_1(char c)
{
	return base64_table[((unsigned)c) & 0xff];
}

int b64_decode(char *dest, const char *src, int len)
{
	int c, b;
	int mod = 0;
	int count = 0;
	const char *p;

	if (!src)
		return -1;
	if (len < 0)
		len = 0;

	for (p = src; (c = b64_decode_1(*p)) != EOS; p++) {
		int x;
		int xc = c;

		if (c == END || c == INV)
			xc = 0;

		switch(mod) {
		case 0:
			x = -1;
			b = xc;
			break;
		case 1:
			x = (b << 2) | (xc >> 4);
			b = xc & 0xf;
			break;
		case 2:
			x = (b << 4) | (xc >> 2);
			b = xc & 0x3;
			break;
		case 3:
			x = (b << 6) | xc;
			b = 0;
			mod = -1;
		}

		if (c == END)
			break;
		if (c == INV) {
			count = -1;
			break;
		}
		if (x >= 0 && count < len)
			dest[count++] = x;
		mod++;
	}
	return count;
}

int b64_space(int len)
{
	return (((len + 3) / 4) * 3);
}

#if BASE64_DEBUG
int main(int argc, char **argv) {
	int len, i, c;
	char *b;

	if (argc < 2) {
		fprintf(stderr, "Usage: %s base64\n", argv[0]);
		return 0;
	}

	len = b64_space(strlen(argv[1]));

	b = malloc(len);
	if (!b) {
		fprintf(stderr, "malloc error\n");
		return 1;
	}
	c = b64_decode(b, argv[1], len);

	printf("decoded = %d\n", c);
	for (i = -1; i < c; i++) {
		putchar(b[i] & 0xff);
	}
	puts("");
}
#endif

